import random
import time

TIMEOUT = 0.2


def solve(people, invalid_links):
    """Determine who gives to whom. `people`: a list of Person objects. `invalid_links`: a dict that defines who
    can't give to whom. Returns a dict. Raises SolvingError if it cannot solve."""
    start_time = time.time()
    random.shuffle(people)
    while not is_valid(dictize(people), invalid_links):
        random.shuffle(people)
        if start_time + TIMEOUT < time.time():
            raise SolvingError('Could not solve.')
    return dictize(people)


def is_valid(people_dict, invalid_links):
    for gifter, giftee in people_dict.items():
        if giftee in invalid_links.get(gifter, ()):
            return False
    return True


def dictize(people_list):
    people = {}
    last_person = people_list[-1]
    for this_person in people_list:
        people[last_person] = this_person
        last_person = this_person
    return people


class SolvingError(Exception):
    pass



class Person:
    """A class to represent a person with an email address."""

    def __init__(self, name, address):
        self.name = name
        self.address = address

    def __repr__(self):
        return '{} <{}>'.format(self.name, self.address)

    def __eq__(self, other):
        return isinstance(other, Person) and self.address == other.address and self.name == other.name

    def __hash__(self):
        # I don't know if this is a valid way to hash and prevent collisions
        return hash(self.name) * hash(self.address)



from .email_util import send_interface
from .person import Person
from .solver import solve


def run(person_list, invalid_links=None, *, message='', subject='Secret Santa'):
    """Solve and email gifters. Raises SolvingError if it cannot solve."""
    if invalid_links is None:
        invalid_links = dict()
    if not isinstance(person_list, list):
        raise ValueError('person_list must be a list')
    if not isinstance(invalid_links, dict):
        raise ValueError('invalid_links must be a dict')
    if not all(isinstance(p, Person) for p in person_list):
        raise ValueError('Every item in person_list must be a Person')

    solved = solve(person_list, invalid_links)  # solve
    failed = send_interface(solved, message, subject)  # send emails (errors are returned)

    if failed:
        print('{} failures: {}'.format(len(failed), failed))



from distutils.core import setup

setup(
    name='santa',
    packages=['santa'],
    version='0.1.1',
    description='Assign a secret santa over email',
    author='jarhill0',
    author_email='jarhill0@gmail.com',
    url='https://github.com/jarhill0/santa/',
    download_url='https://github.com/jarhill0/santa/archive/0.1.1.tar.gz',
    keywords=['secret', 'santa', 'email'],
    classifiers=[],
    install_requires=[
        'yagmail >=0.10.209',
    ]
)



import yagmail

yag = yagmail.SMTP()


def _send(address, gifter_name, giftee_name, message, subject):
    """Format and send a message to alert a participant about their assigned recipient."""
    body = 'Hi {},\n\n'.format(gifter_name) + message.strip() + '\n\nYour assigned recipient is {}.'.format(giftee_name)
    yag.send(to=address, subject=subject, contents=body)


def send_interface(solved_dict, message, subject):
    failed = []
    for gifter, recipient in solved_dict.items():
        add = gifter.address
        g_name = gifter.name
        r_name = recipient.name
        try:
            _send(add, g_name, r_name, message, subject)
        except Exception:
            failed.append((gifter, recipient))
    return failed
